#include <stdio.h>
#include <iodef.h>
#include <brkdef.h>
#include <descrip.h>
#include <ssdef.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "global.h"
#include "constants.h"
#include "window.h"
#include "extern.h"
#include "qio.h"

#define lf '\012'                    /* Terminator for a line in irc */

static unsigned short port = DEF_PORT;      /* The standard IRC port. */

extern int sys$setast(), sys$clref(),
       sys$cantim(), lib$get_ef(), sys$bintim();
extern long myaddress;
extern win_ptr awins;
extern int setup_qio_hooks(char *);
extern unsigned *kill_server_channels(unsigned *, char **);
extern void win_serverset(server_ptr srv);

int quit = FALSE;
static char ast_wait[] = "0000 00:00:00.30";
static $DESCRIPTOR(ast_wait_d, ast_wait);
static unsigned ast_time[2];
void irc_receive();
char def_nick[MAXLEN];

server_ptr server_list = NULL;
extern server_ptr gsrv;
extern int print_window();
static void process_receive();

/* ------------------------------------------------------------------------ */
void set_redirect(char *str, server_ptr srv)
{
  if (!srv) return;
  if (srv->redirect) myfree(srv->redirect);
  if (str) srv->redirect = m_strcpy(str);
  else srv->redirect = NULL;
}

/* ------------------------------------------------------------------------ */
server_ptr lookup_server(char *server, char *nick, int connected, int port)
{
  server_ptr walk;
  for (walk=server_list;walk;walk=walk->next) {
    if ((strcasecmp(server, walk->server_name)==0) &&
        (!nick || !walk->nick || (strcasecmp(nick, walk->nick) == 0)) &&
        (!port || (walk->port == port)))
      if (connected == walk->connected) return walk;
  }
  return NULL;
}

/* ------------------------------------------------------------------------ */
static void server_disconnected(server_ptr serv)
{
  win_ptr walk;
  if (serv->connected) {
    if (serv->to_join) myfree(serv->to_join);
    serv->chan_list = kill_server_channels(serv->chan_list, &serv->to_join);
  }
  tcp_shutdown(serv->iochan);
  serv->iochan = 0;
  myfree(serv->redirect);
  serv->redirect = NULL;
  if (serv->connected) (void)handle_on(o_disconnect, serv->server_name, 0);
  serv->away = serv->connected = FALSE;
}

/* ------------------------------------------------------------------------ */
static int scan_server_list(short chan)
{
  server_ptr temp;
  unsigned status;
  int old;

  temp = server_list;

  while (temp) {
    if (sys$clref(temp->event_flag) == SS$_WASSET)
      if (temp->connected) 
        process_receive(temp);
      else {
        server_disconnected(temp);
      }
    temp = temp->next;
  }
  return 1;
}

/* ------------------------------------------------------------------------ */
extern unsigned lib$free_ef();
/* ------------------------------------------------------------------------ */
void remove_server(server_ptr srv, int reason)
{
  if (!srv) return;
  if (!srv->connected) return;
  if (reason==1) new_send(srv, "QUIT :No more windows left!\n");
  else if (reason==2) new_send(srv, "QUIT :Switching servers!\n");
  else if (reason==3) ;
  srv->connected = FALSE;
  if (srv->to_join) myfree(srv->to_join);
  srv->chan_list = kill_server_channels(srv->chan_list, &srv->to_join);
}
/* ------------------------------------------------------------------------ */
static int my_nslookup(char *host, char *add)
{
  char *local;
  if (good_host(add, host)) return 1;
  local = getvar("NAMESERVER");
  if (local && nslookup(local, host, add)) return 1;
  if (nslookup("128.205.1.2", host, add)) return 1;
  if (nslookup("128.205.7.1", host, add)) return 1;
  say("*** Fatal error in nslookup of %s", host);
  return 0;
}

/* ------------------------------------------------------------------------ */
short open_connection(server, port)
  char *server;
  unsigned short port;
{
  char msg[80], *ns_local;
  short serv_chan;
  struct sockaddr_in_type sin;         /* The sin for our connection */

  sin.sin_family = AF_INET;
  sin.sin_port = tcp_htons(port);
  sin.sin_zero[0] = 0;
  sin.sin_zero[1] = 0;
  say("*** Connecting to port %d of %s", port, server);
  if (!my_nslookup(server, (char *)&sin.sin_address)) return -1;
  if ((serv_chan = tcp_socket (AF_INET, SOCK_STREAM, PF_INET)) == -1) {
    get_socket_error(msg);
    say("socket: %s", msg);
    return -1;
  }
  if (tcp_connect(serv_chan, &sin, sizeof(sin)) == -1) {
    get_socket_error(msg);
    say("connect: %s", msg);
    return -1;
  }
  return serv_chan;
}

/* ------------------------------------------------------------------------ */
static server_ptr insert_server(server_ptr top, server_ptr new)
{
  server_ptr walk;
  for (walk = top; walk && walk->next; walk = walk->next) ;
  new->next = NULL;
  if (walk) walk->next = new;
  else top = new;
  return top;
}
/* ------------------------------------------------------------------------ */
static void init_server(server_ptr temp)
{
/* Fields not set: server_name port nick chan_list notify_list last_notify
                   next
*/
  temp->iochan = 0;
  temp->event_flag = 0;
  if (temp->redirect) myfree(temp->redirect);
  memset(&(temp->iosb), 0, sizeof(temp->iosb));
  temp->redirect = temp->last = NULL;
  temp->buffer[0] = temp->start[0] = 0;
  temp->connected = FALSE;
  temp->away = FALSE;
}
/* ------------------------------------------------------------------------ */
static server_ptr new_server_node(char *serv, char *nick, char *pass, int port)
{
  server_ptr temp;
  temp = (server_ptr)mymalloc(sizeof(struct server));
  if (!temp) return NULL;
  memset(temp, 0, sizeof(struct server));
  strcpy(temp->server_name, serv);
  if (nick) temp->nick = m_strcpy(nick);
  else temp->nick = NULL;
  temp->port = port?port:DEF_PORT;
  if (pass) temp->pass = m_strcpy(pass);
  else temp->pass = NULL;
  init_server(temp);
  server_list = insert_server(server_list, temp);
  return temp;
}
/* ------------------------------------------------------------------------ */
static void fillin_server_values(server_ptr ptr, server_ptr old)
{
  if (!ptr) return;
  if (!ptr->nick || !*ptr->nick)
    if (old && old->nick) ptr->nick = m_strcpy(old->nick);
    else if (gsrv && gsrv->nick) ptr->nick = m_strcpy(gsrv->nick);
    else ptr->nick = m_strcpy(def_nick);
  if (!ptr->port) ptr->port = DEF_PORT;
}

extern notify_ptr copy_notify_list(notify_ptr);
/* ------------------------------------------------------------------------ */
/* If old is NULL, then we are changing the server associated with 1 window
/* If old is !NULL, then we are changing all of instances of the old
/*   server */
static int connect_server_by_ptr(server_ptr ptr, server_ptr old)
{
  int iochan, status;
  win_ptr walk;

  fillin_server_values(ptr, old);
  iochan = open_connection(ptr->server_name, ptr->port);
  if (iochan == -1) return FALSE;
  ptr->iochan = iochan;
  if (!(status = lib$get_ef(&(ptr->event_flag)))) {
    say("*** Could not allocated event flag for server.");
    return FALSE;
  }
  sys$clref(ptr->event_flag);
  if (!ptr->notify_list)
    if (old) ptr->notify_list = copy_notify_list(old->notify_list);
    else if (gsrv) ptr->notify_list = copy_notify_list(gsrv->notify_list);
    else ptr->notify_list = copy_notify_list(NULL);
  if (old) { /* had a check to see if it was connected.. why? */
    remove_server(old, 2);
    if (old != ptr) {
      for (walk=awins; walk; walk = walk->next)
        if (walk->server == old) walk->server = ptr;
      if (old->to_join) {
        if (ptr->to_join) myfree(ptr->to_join);
        ptr->to_join = m_strcpy(old->to_join);
      }
    }
  }
/* This chunk of code will set any window that doesn't already have a server */
  for (walk=awins; walk; walk = walk->next)
    if (walk->server == NULL) walk->server = ptr;
  ptr->connected = TRUE;
  win_serverset(ptr);
  startup(ptr, ptr->pass);
  if (!handle_on(o_connect, ptr->server_name, 0))
    say("*** New server set.");
  irc_receive(ptr);
  return TRUE;
}
/* ------------------------------------------------------------------------ */
void connect_server_by_number(int num)
{
  server_ptr list = server_list;
  while (num-->0) list = list->next;
  if (list)
    if (!list->connected) connect_server_by_ptr(list, gsrv);
    else say("*** Already connected to that server");
  else say("*** There aren't that many servers!");
}
/* ------------------------------------------------------------------------ */
int *add_server(server, nick, port, pass, oldserv, match)
  char *server, *pass, *nick;
  unsigned short port;
  server_ptr oldserv, match;
{
  server_ptr temp;

  if (!match) {
    if (!(temp = new_server_node(server, nick, pass, port))) {
      say("*** Could not malloc space for server.");
      return NULL;
    }
  } else {
    temp = match;
    temp->port = port;
    myfree(temp->nick);
    temp->nick = m_strcpy(nick);
    init_server(temp);
  }
  temp->last_notify = NULL;
  temp->chan_list = NULL;

  connect_server_by_ptr(temp, oldserv);
  return (int *)temp;
}

/* ------------------------------------------------------------------------ */
extern int waiting;
static char *parse_buffer(char *str, server_ptr srv)
{
  char *dummy, good = TRUE;
  dummy = str;
  do {
    while (*str && (*str != lf)) str++;
    if (*str==lf) {
      *str = 0;
      if (str[-1] == 13) str[-1] = 0;
      handle_line(dummy, srv);
      dummy = ++str;
    } else good = FALSE;
  } while (good&&waiting);
  return dummy;
}

/* ------------------------------------------------------------------------ */
void process_wait(server_ptr gsrv)
{
  char tmp[MAXLEN], save[MAXLEN*2], *ret = save;
  struct iosb_type iosb;
  int num, good;
  save[0] = 0;
/*  yell("[%s] still stuck in buffer", gsrv->last); */
  strcpy(save, gsrv->last);
  do {
    num = tcp_receive(gsrv->iochan, tmp, sizeof(tmp)-1, &iosb, 0, NULL);
    good = 1;
    if (iosb.length>0) {
      ret = save;
      tmp[iosb.length] = 0;
      memcpy(&save[strlen(save)], tmp, iosb.length+1);
      ret = parse_buffer(save, gsrv);
      if (save!=ret) strcpy(save, ret);
    }
    say("waiting = %d", waiting);
  } while (waiting && (iosb.length > 0));
  strcpy(gsrv->last+1, ret);
/*  yell("leaving [%s] in buffer", ret); */
}

/* ------------------------------------------------------------------------ */
static void process_receive(temp)
  server_ptr temp;
{
  int good;
  char *dummy, *walk;

  if (quit) return;
  if (((int)temp->iosb.length > 0) && ODD(temp->iosb.status))
  {
    temp->buffer[(int)temp->iosb.length] = 0;
    memcpy(&temp->start[strlen(temp->start)], temp->buffer, temp->iosb.length+1);
    walk = dummy = temp->start;
    good = TRUE;
    do {
      while (*walk && (*walk != lf)) walk++;
      if (*walk==lf) {
        *walk = 0;
        if (walk[-1] == 13) walk[-1] = 0;
        temp->last = walk+1;
        handle_line(dummy, temp);
        dummy = ++walk;
      } else good = FALSE;
    } while (good);
    (void)strcpy(temp->start, dummy);
    irc_receive(temp);
    return;
  }
  say("The connection has been closed by the server.");
  server_disconnected(temp);
}

/* ------------------------------------------------------------------------ */
/* The order of things:
 *   1. add_connect_to_list() to insert a connection into the queue
 *   2. start_read() to let the master_read function know to start
 *      monitoring the connection (and tell it what to do with the data)
/* ------------------------------------------------------------------------ */
static connect_node *master_connect_list = NULL;
/* ------------------------------------------------------------------------ */
void master_list(char *str)
{
  int old;
  connect_node *w;
  old = sys$setast(0);
  say("*** T chan acce host_id   port [%s]", str);
  for (w = master_connect_list; w; w = w->next)
  {
    say("*** %1.1d %4.4d %4.4d %9.9u %5.5d", w->type, w->iochan,
        w->accept_chan, w->sin?tcp_htonl(w->sin->sin_address):-1,
        w->sin?tcp_htons(w->sin->sin_port):-1);
  }
  say("*** End of connect list");
  if (old==SS$_WASSET) sys$setast(1);
}

/* ------------------------------------------------------------------------ */
static int master_read(short chan)
{              
  int old, kill_it;
  connect_node *walk, *save, *p = NULL;
  old = sys$setast(0);
  walk = master_connect_list;
  while (walk)
    if ((sys$clref(walk->event_flag) == SS$_WASSET) &&
        (walk->iosb.status != 0))
    {
      kill_it = FALSE;
      if ((walk->iosb.status != 1) || (walk->iosb.length <= 0))
      {
        if (walk->on_exit) walk->on_exit(walk);
        else say("death %d %d", walk->iosb.status, walk->iosb.length);
        kill_it = TRUE;
      } else {
        if (walk->on_read)
          if (walk->on_read(walk))
            tcp_receive(walk->iochan, walk->buffer, sizeof(walk->buffer),
                        &walk->iosb, walk->event_flag, master_read);
          else
            kill_it = TRUE;
        else say("%s", walk->buffer);
      }
      if (kill_it) {
        tcp_shutdown(walk->iochan);
        lib$free_ef(&walk->event_flag);
        if (walk->sin) myfree(walk->sin);
        if (walk->specific) myfree(walk->specific);
        if (p) p->next = walk->next;
        else master_connect_list = walk->next;
        save = walk;
        walk = walk->next;
        myfree(save);
      } else { p = walk; walk = walk->next; }
    } else { p = walk; walk = walk->next; }
  if (old==SS$_WASSET) sys$setast(1);
  return 1;
}

/* ------------------------------------------------------------------------ */
char *start_read(short chan, short nchan, READ_ROUT on_read, READ_ROUT on_exit)
{
  int old;
  connect_node *walk;
  old = sys$setast(0);
  for (walk = master_connect_list; walk; walk = walk->next)
    if (walk->iochan == chan)
    {
      if (walk->iochan != nchan) {
        tcp_shutdown(walk->iochan);
        walk->iochan = nchan;
      }
      walk->on_read = on_read;
      walk->on_exit = on_exit;
      tcp_receive(walk->iochan, walk->buffer, sizeof(walk->buffer),
                  &walk->iosb, walk->event_flag, master_read);
      if (old==SS$_WASSET) sys$setast(1);
      return walk->specific;
    }
  if (old==SS$_WASSET) sys$setast(1);
  return NULL;
}

/* ------------------------------------------------------------------------ */
void add_connect_to_list(short chan, SIN *mysin, char *specific, int type)
{
  int status;
  connect_node *temp;
  temp = (connect_node *)mymalloc(sizeof(connect_node));
  if (!temp) return;
  temp->type = type;
  temp->iochan = chan;
  temp->sin = mysin;
  temp->buffer[0] = 0;
  status = lib$get_ef(&(temp->event_flag));
  if (!(status & 1)) temp->event_flag = 0;
  temp->on_read = temp->on_exit = NULL;
  temp->specific = specific;
  temp->next = master_connect_list;
  master_connect_list = temp;
}

/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
int new_send(server_ptr node, char *format, ...)
{
  int ss;
  char string[MAXLEN*2];
  va_list args;

  if (!node || !node->connected) {
    say("*** You are not connect to a server - try /server %s",
        node?node->server_name:"irc.colorado.edu");
    return 0;
  }
  va_start(args, format);
  vsprintf(string, format, args);
  va_end(args);
  if (getivar("DEBUG") & DEBUG$SEND)
    yell("%s", string);

  if ((ss = tcp_send(node->iochan, string, strlen(string))) == -1) {
    get_socket_error(string);
    say("send: %s", string);
    server_disconnected(node);
  }
  return ss;
}

/* ------------------------------------------------------------------------ */
void print_server_list()
{
  server_ptr s;
  int count = 0;
  say("*** Server list:");
  for (s = server_list; s; s = s->next, count++)
    if (s->nick)
      say("***   %2d) %s %d (%s%s) [%d]", count, s->server_name, s->port,
          s->connected?"":"was ", s->nick, s->iochan);
    else
      say("***   %2d) %s %d [%d]", count, s->server_name, s->port, s->iochan);
}

/* ------------------------------------------------------------------------ */
void irc_receive(temp)
  server_ptr temp;
{
  int ss;
  if (!temp) return;
  ss = tcp_receive(temp->iochan, temp->buffer, sizeof(temp->buffer)-1,
                   &(temp->iosb), temp->event_flag, scan_server_list);
  if (ss==-1) {
    char msg[80];
    get_socket_error(msg);
    say("irc_receive: %s", msg);
  }
}

/* ------------------------------------------------------------------------ */
int good_host(int_addr, host)
  unsigned char int_addr[4];
  char *host;
{
  int ack;
  ack = sscanf(host, "%d.%d.%d.%d", &int_addr[0], &int_addr[1], &int_addr[2],
               &int_addr[3]);
  if (ack==4) return TRUE;
  return FALSE;
}

/* ------------------------------------------------------------------------ */
static void getmy_host(void)
{
  char hostname[80], *host;
  host = getvar("HOSTNAME");
  if (!host || !*host) {
    tcp_gethostname(hostname);
    host = hostname;
  }
  if (host && *host)
    if (!my_nslookup(host, (char *)&myaddress))
    {
      say("*** Fatal error in nslookup of %s. Aborting", host);
      say("*** You will not be able to dcc send");
    }
   else say("*** %s translated.", host);
}

/* ------------------------------------------------------------------------ */
extern void free_hold();
/* ------------------------------------------------------------------------ */
static void interactive(void)
{
  char input[MAXLEN], *walk, line[MAXLEN];

  while (1) {
    walk = input;
    (void)read_input(input, 560, NULL);
    if (*input) {
      if (getbvar("INPUT_ALIASES")) 
        parse_line(input, "", FALSE, TRUE);
      else
        parse_line(input, NULL, FALSE, TRUE);
      add_input_list(input);
    }
    free_hold();
    updatestatwin();
    find_input_prompt(FALSE);
  }
}

/* ------------------------------------------------------------------------ */
static void error_handler(void)
{
  quit = TRUE;
  delete_windows();
  cleanup();
  (void)printf("This is really lame\n");
}

/* ------------------------------------------------------------------------ */
extern int DUMB;
extern void lib$establish();
extern char global_ircrc[MAXLEN];

/* ------------------------------------------------------------------------ */
static void read_servers_file(char *file)
{
  FILE *fp;
  char *word, str[MAXLEN], *walk;
  if (!(fp = fopen(file, "r"))) return;
  while (fgets(str, sizeof(str)-1, fp))
  {
    walk = str;
    word = next_arg(&walk);
    new_server_node(word, NULL, NULL, atoi(walk));
  }
  fclose(fp);
}

/* ------------------------------------------------------------------------ */
void master_init(char *gfile, char *pfile)
{
  if (!setup_qio_hooks("disk$userdisk1:[mas0.masdough.irc]"))
  {
    say("Could not setup qio hooks -- aborting");
    exit(0);
  }
  init_binds();
  init_input_list(getivar("HISTORY"));
  load_file(gfile); /* Load the global file */
  getmy_host();
  init_keybdio();
  init_window();    /* dependant on INPUT_DEFINED and DUMB */
  load_file(pfile); /* Load the personal file */
}

/* ------------------------------------------------------------------------ */
void start_first_server()
{
  int status;
  server_ptr temp;
  char start_line[255];
  extern void user_server(char *);

  status = sys$bintim(&ast_wait_d, ast_time);
  if (!ODD(status)) my_sig(status, "ast bintime");
  for (temp = server_list; !gsrv && temp; temp = temp->next)
  {
    sprintf(start_line, "%s:%s:%d:%s", temp->server_name,
            temp->nick?temp->nick:def_nick, temp->port,
            temp->pass?temp->pass:"");
    user_server(start_line);
  }
}

/* ------------------------------------------------------------------------ */
static int old_main(argc, argv)
  int argc;
  char **argv;
{
  unsigned status, dummy;
  char server[MAXLEN], loadfile[80], *ods = NULL, *ds = NULL, base_path[80];
  int curr = 0, port = DEF_PORT, set_server = 0, loop;

  strcpy(base_path, argv[0]);
  ds = base_path;
/* will not work with >'s for directory specifications.
   What if it's a logical: irclib:irc.exe ?? */
  while ((ds = (char *)strstr(ds, "]"))) ods = ++ds;
  if (ods) *ods = 0;
  sprintf(global_ircrc, "%sglobal.ircrc", base_path);

  if (!get_logical("IRCNICK", def_nick, FALSE))
    (void)get_username(def_nick, FALSE);
  if (!get_logical("IRCRC", loadfile, FALSE))
    strcpy(loadfile, "sys$login:.ircrc");
  if (get_logical("IRCSERVER", server, FALSE))
    set_server = 1;

  for (loop=1;loop<argc;loop++) {
    if (argv[loop][0] == '-') {
      if (argv[loop][1] == 'd') DUMB = 1;
      else if (argv[loop][1] == 'b') DUMB = 2;
      else if (argv[loop][1] == 'p') {
         if (++loop<argc) port = atoi(argv[loop]);
      }
      else if (argv[loop][1] == 'l') {
         if (++loop<argc) strcpy(loadfile, argv[loop]);
      }
    } else {
      if (curr==0) (void)strcpy(def_nick, argv[loop]);
      if (curr==1) {
        (void)strcpy(server, argv[loop]);
        set_server = 1;
      }
      curr++;
    }
  }
  master_init(global_ircrc, loadfile);
#ifdef ENABLE_SERVER
  if (set_server)
    new_server_node(server, NULL, NULL, port);
  else
#else
  say("*** The SERVER command has been disabled.");
#endif
  {
    sprintf(global_ircrc, "%sircii.servers", base_path);
    read_servers_file(global_ircrc);
  }

  start_first_server();
  (void)launch_timer(); /* 15 second timer - async operations */
  interactive();        /* Handle user input - never exits */
  return 1;
}

int main(int argc, char *argv[])
{
  lib$establish(error_handler); /* use this so we can read the traceback */
  return old_main(argc, argv);
}
